// lobby.js@0.0.0nr

// Makes it easier to force a string length for buffering.
const reqstr = require('./require-strlen');
const sanitorium = require('./maps/sanitorium');

// Function for shuffling arrays.
function shuffle(array) {
  var currentIndex = array.length, temporaryValue, randomIndex;

  // While there remain elements to shuffle...
  while (0 !== currentIndex) {

    // Pick a remaining element...
    randomIndex = Math.floor(Math.random() * currentIndex);
    currentIndex -= 1;

    // And swap it with the current element.
    temporaryValue = array[currentIndex];
    array[currentIndex] = array[randomIndex];
    array[randomIndex] = temporaryValue;
  }

  return array;
}

/** This constructor creates a game lobby. */
module.exports = class Lobby {
  // The primary constructor.
  constructor(loggingId) {
    this.dc = false;
    this.maps = [];
    this.maps.push(sanitorium());
    this.collected = 0;
    this.id = loggingId;
    this.locked = false;
    this.map = this.maps[Math.floor(Math.random() * this.maps.length)];
    this.over = false;
    this.players = [];
  }

  checkWinningConditions() {
    // Game Page Winning Conditions
    let playersAlive = 0;
    this.players.forEach((ply) => {
      if (ply.role === 'survivor' && ply.killed === false) {
        playersAlive += 1;
      }
    });

    if (this.dc === true) {
      this.players.forEach((ply) => {
        ply.write(ply.role === 'killer' ? 't' : 'y');
        ply.destroy();
      });
      this.over = true;
      return true;
    }

    if (playersAlive <= 0) {
      this.players.forEach((ply) => {
        ply.write(ply.role === 'killer' ? 'k' : 'd');
        ply.destroy();
      });
      this.over = true;
      return true;
    }

    // Game Page Winning Conditions
    if (this.map.pages.length <= 0) {
      this.players.forEach((ply) => {
        ply.write(ply.role === 'killer' ? 'l' : 'w');
        ply.destroy();
      });
      this.over = true;
      return true;
    }
  }

  /** Initializes the lobby. */
  initialize() {
    // Makes sure the game can't start with less than 2 players.
    if (this.players.length <= 1) { return; }
    this.players = shuffle(this.players);

    // Setup all of the players.
    this.players.forEach((player, index) => {
      // Player Information
      player.id = index;
      player.killed = false;
      player.lobbyId = this.id;
      player.queued = false;
      player.queuedAction = [];

      // Handles killer randomization.
      if (index === 0) {
        player.role = 'killer';
      } else {
        player.role = 'survivor';
      }

      // Player Position Data
      if (player.role === 'killer') {
        let randomSpawn = Math.floor(Math.random() * this.map.spawns.killer.length);
        player.x = this.map.spawns.killer[randomSpawn][0];
        player.y = this.map.spawns.killer[randomSpawn][1];
        this.map.spawns.killer.splice(randomSpawn, 1);
      } else {
        let randomSpawn = Math.floor(Math.random() * this.map.spawns.survivor.length);
        player.x = this.map.spawns.survivor[randomSpawn][0];
        player.y = this.map.spawns.survivor[randomSpawn][1];
        this.map.spawns.survivor.splice(randomSpawn, 1);
      }
    });

    // Locks the lobby from new players.
    this.locked = true;

    // Tells all of the clients that the server is ready for inputs.
    this.players.forEach((player) => {
      player.write('r');
    });

    // Render the map of everyone.
    this.renderMaps();
  }

  /** Adds a player to the lobby. It's recommended not to use this mid-game. */
  insertPlayer(socket) {
    // Prevents a player from joining a locked lobby.
    if (this.locked === true) { return; }

    // Adds the player in the list.
    this.players.push(socket);
  }

  /** Queues a player's movement for the next step. */
  queueMovement(socket, action) {
    // Setup a player's queue.
    if (action === 'w') {
      if (socket.y - 1 < 0) { return; }
      if (socket.y - 1 > this.map.map.length) { return; }
      if (this.map.objects.includes(this.map.map[socket.y - 1][socket.x])) { return; }
      if (this.map.kObjects.includes(this.map.map[socket.y - 1][socket.x]) && socket.role === 'survivor') { return; }
      socket.queuedAction = [socket.x, socket.y - 1];
      socket.queued = true;
    }
    
    if (action === 'x') {
      if (socket.y + 1 < 0) { return; }
      if (socket.y + 1 > this.map.map.length) { return; }
      if (this.map.objects.includes(this.map.map[socket.y + 1][socket.x])) { return; }
      if (this.map.kObjects.includes(this.map.map[socket.y + 1][socket.x]) && socket.role === 'survivor') { return; }
      socket.queuedAction = [socket.x, socket.y + 1];
      socket.queued = true;
    }
    
    if (action === 'a') {
      if (socket.x - 1 < 0) { return; }
      if (socket.x - 1 > this.map.map[socket.y].length) { return; }
      if (this.map.objects.includes(this.map.map[socket.y][socket.x - 1])) { return; }
      if (this.map.kObjects.includes(this.map.map[socket.y][socket.x - 1]) && socket.role === 'survivor') { return; }
      socket.queuedAction = [socket.x - 1, socket.y];
      socket.queued = true;
    }
    
    if (action === 'd') {
      if (socket.x + 1 < 0) { return; }
      if (socket.x + 1 > this.map.map[socket.y].length) { return; }
      if (this.map.objects.includes(this.map.map[socket.y][socket.x + 1])) { return; }
      if (this.map.kObjects.includes(this.map.map[socket.y][socket.x + 1]) && socket.role === 'survivor') { return; }
      socket.queuedAction = [socket.x + 1, socket.y];
      socket.queued = true;
    }

    if (action === 's') {
      socket.queuedAction = [socket.x, socket.y];
      socket.queued = true;
    }

    if (action === 'q') {
      if (socket.role === 'survivor') { return; }
      if (socket.x - 1 < 0) { return; }
      if (socket.x - 1 > this.map.map[socket.y - 1].length) { return; }
      if (socket.y - 1 < 0) { return; }
      if (socket.y - 1 > this.map.map.length) { return; }
      if (this.map.objects.includes(this.map.map[socket.y - 1][socket.x - 1])) { return; }
      socket.queuedAction = [socket.x - 1, socket.y - 1];
      socket.queued = true;
    }

    if (action === 'e') {
      if (socket.role === 'survivor') { return; }
      if (socket.x + 1 < 0) { return; }
      if (socket.x + 1 > this.map.map[socket.y - 1].length) { return; }
      if (socket.y - 1 < 0) { return; }
      if (socket.y - 1 > this.map.map.length) { return; }
      if (this.map.objects.includes(this.map.map[socket.y - 1][socket.x + 1])) { return; }
      socket.queuedAction = [socket.x + 1, socket.y - 1];
      socket.queued = true;
    }

    if (action === 'z') {
      if (socket.role === 'survivor') { return; }
      if (socket.x - 1 < 0) { return; }
      if (socket.x - 1 > this.map.map[socket.y + 1].length) { return; }
      if (socket.y + 1 < 0) { return; }
      if (socket.y + 1 > this.map.map.length) { return; }
      if (this.map.objects.includes(this.map.map[socket.y + 1][socket.x - 1])) { return; }
      socket.queuedAction = [socket.x - 1, socket.y + 1];
      socket.queued = true;
    }

    if (action === 'c') {
      if (socket.role === 'survivor') { return; }
      if (socket.x + 1 < 0) { return; }
      if (socket.x + 1 > this.map.map[socket.y + 1].length) { return; }
      if (socket.y + 1 < 0) { return; }
      if (socket.y + 1 > this.map.map.length) { return; }
      if (this.map.objects.includes(this.map.map[socket.y + 1][socket.x + 1])) { return; }
      socket.queuedAction = [socket.x + 1, socket.y + 1];
      socket.queued = true;
    }

    // Deal with checking queues for step logic.
    let queuedPlayers = 0;
    let deadPlayers = 0;
    for (let i = 0; i < this.players.length; i += 1) {
      if (this.players[i].killed === true) { deadPlayers += 1; }
      if (this.players[i].queued === false) { continue; }
      queuedPlayers += 1;
    }

    /** Check if all players are ready for a movement step. */
    if (queuedPlayers === this.players.length - deadPlayers) {
      // Move all players.
      this.players.forEach((player) => {
        // Set the player's coordinates.
        player.x = player.queuedAction[0];
        player.y = player.queuedAction[1];
        player.queuedAction = [];
        player.queued = false;
      });

      this.players.forEach((player) => {
        // Checks if the player is killing someone.
        if (player.role === 'killer') {
          this.players.forEach((victim) => {
            if (victim === player) { return; }
            if (victim.x === player.x && victim.y === player.y) {
              victim.write('d');
              victim.killed = true;
              victim.destroy();
            }
          });
        }

        // Checks if the player is walking into the killer.
        if (player.role === 'survivor') {
          this.players.forEach((killer) => {
            if (killer.role === 'survivor') { return; }
            if (killer.x === player.x && killer.y === player.y) {
              player.write('d');
              player.killed = true;
              player.destroy();
            }
          });
        }

        if (player.role === 'survivor') {
          this.map.pages.forEach((page, ind) => {
            if (page[0] === player.x && page[1] === player.y) {
              this.map.pages.splice(ind, 1);
            }
          });
        }

        if (this.checkWinningConditions() === true) { return; };
        this.renderMaps();
      });
    }
  }

  renderMap(player) {
    // Update all the player's games.
      let parsedMap = '';
      if (player.role === 'killer') {
        // 9x9 board
        for (let i = player.y - 9; i <= player.y + 9; i += 1) {
          for (let j = player.x - 9; j <= player.x + 9; j++) {
            parsedMap += this.map.map[i][j];
          }
          parsedMap += '\n';
        }
      } else {
        // 5x5 board
        for (let i = player.y - 5; i <= player.y + 5; i += 1) {
          for (let j = player.x - 5; j <= player.x + 5; j++) {
            parsedMap += this.map.map[i][j];
          }
          parsedMap += '\n';
        }
      }

      // Return the player's map.
      return parsedMap;
  }

  renderMaps() {
    // Render all of the players' maps.
    this.players.forEach((player) => {
      if (player.killed === true) { return; }
      const playerMap = this.renderMap(player);
      const playerPages = this.renderPages(player);
      const playerPlayers = this.renderPlayers(player);
      player.write(`u${reqstr(playerMap.length, 9, '0', true)}${playerMap}${playerPlayers}${playerPages}`);
    });
  }

  renderPages(player) {
    // Page Variables
    let pageCount = 0;
    let renderedPages = '';

    // Go through all of the pages.
    this.map.pages.forEach((page) => {
      if (player.role === 'survivor') {
        if (page[0] - player.x < -5 || page[0] - player.x > 5) { return; }
        if (page[1] - player.y < -5 || page[1] - player.y > 5) { return; }
        renderedPages += reqstr((page[0] - player.x) + 5, 3, '0', true);
        renderedPages += reqstr((page[1] - player.y) + 5, 3, '0', true);
        pageCount += 1;
      } else {
        pageCount = 0;
      };
    });

    // Return all of the important data.
    renderedPages = reqstr(pageCount, 1, '0', true) + renderedPages;
    return renderedPages;
  }

  renderPlayers(player) {
    let playerCount = 1;
    let renderedMaps = '';
    this.players.forEach((ply) => {
      if (player === ply) {
        if (player.role === 'killer') {
          renderedMaps += reqstr(9, 3, '0', true);
          renderedMaps += reqstr(9, 3, '0', true);
          renderedMaps += '?';
        } else {
          renderedMaps += reqstr(5, 3, '0', true);
          renderedMaps += reqstr(5, 3, '0', true);
          renderedMaps += '@';
        }
      } else {
        if (ply.killed === true) { return; } 
        if (player.role === 'killer') {
          if (ply.x - player.x < -9 || ply.x - player.x > 9) { return; }
          if (ply.y - player.y < -9 || ply.y - player.y > 9) { return; }
          renderedMaps += reqstr((ply.x - player.x) + 9, 3, '0', true);
          renderedMaps += reqstr((ply.y - player.y) + 9, 3, '0', true);
          if (ply.role === 'killer') {
            renderedMaps += '?';
          } else {
            renderedMaps += '@';
          }
          playerCount += 1;
        } else {
          if (ply.x - player.x < -5 || ply.x - player.x > 5) { return; }
          if (ply.y - player.y < -5 || ply.y - player.y > 5) { return; }
          renderedMaps += reqstr((ply.x - player.x) + 5, 3, '0', true);
          renderedMaps += reqstr((ply.y - player.y) + 5, 3, '0', true);
          if (ply.role === 'killer') {
            renderedMaps += '?';
          } else {
            renderedMaps += '@';
          }
          playerCount += 1;
        }
      };
    });

    // Return Data
    renderedMaps = reqstr(playerCount, 1, '0', true) + renderedMaps;
    return renderedMaps;
  }

  /** Removes a player from the game. ONLY DO THIS WHEN SOMEONE DISCONNECTS. */
  removePlayer(socket) {
    if (typeof this.players[socket] === 'object') {
      this.players.splice(socket, 1);
    };
  }
};
